#!/usr/bin/env python
#
# GEemyle: Eemyle Connector using XML-RPC
# Copyright (C) 2003-2006 Easter-eggs <info@easter-eggs.com>
# Authors:
#      Cyril Lacoux (Yack) <clacoux@easter-eggs.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

__author__  = "Cyril Lacoux [Yack] <clacoux@easter-eggs.com>"
__cvsid__   = "$Id$"
__version__ = "$Revision$"[11:-2]
__all__     = [ "EemyleConnector"]

import threading
import gobject
import urllib
import xmlrpclib
import string
import time
import types
import sys

marshaller = xmlrpclib.Marshaller(allow_none=True)

def dump_struct(dict):
    """
    Simple function to convert a dictionary to a string
    """
    global marshaller

    out   = []
    write = out.append

    marshaller.dump_struct(dict, write)

    return string.join(out, "")

class EemyleConnectorError(xmlrpclib.Error):

    def __init__(self):

        xmlrpclib.Error.__init__(self)
        self.errmsg  = 'Unknown error'
        self.errcode = -1

    def __repr__(self):
        return "<EemyleConnectorError: %d %s>" % (self.errmsg, self.errcode)


class EemyleConnector(threading.Thread, gobject.GObject):
    """
    XML-RPC Connector to Eemyle server
    """

    def __init__(self, hostname='', wservice='', username='', password='', scheme='http', port='80', verbose=False):
        """
        Initializing some stuff
        """

        threading.Thread.__init__(self, name='Eemyle XML-RPC Connector')
        gobject.GObject.__init__(self)

        self.__hostname = hostname
        self.__wservice = wservice
        self.__username = username
        self.__password = password
        self.__scheme   = scheme
        self.__port     = port
        self.__verbose  = verbose

        # Internal vars
        self.__server       = None
        self.__do_run       = True
        self.__request_fifo = []

        self.__build_server()

        # Start the thread
        self.start()

    def set_hostname(self, hostname):
        """
        Change the hostname
        """

        if hostname == self.__hostname:
            return True

        self.__hostname = hostname
        self.__build_server()

    def set_wservice(self, wservice):
        """
        Change the web service
        """

        if wservice == self.__wservice:
            return True

        self.__wservice = wservice
        self.__build_server()

    def set_username(self, username):
        """
        Change the username
        """

        if username == self.__username:
            return True

        self.__username = username
        self.__build_server()

    def set_password(self, password):
        """
        Change the password
        """

        if password == self.__password:
            return True

        self.__password = password
        self.__build_server()

    def set_scheme(self, scheme):
        """
        Change the scheme
        """
        
        if scheme == self.__scheme:
            return True

        self.__scheme = scheme
        self.__build_server()

    def set_port(self, port):
        """
        Change the port
        """
        
        if port == self.__port:
            return True

        self.__port = port
        self.__build_server()

    def __build_server(self):
        """
        Builds the server depending the connection parameters
        """

        # Constructing the URI 
        scheme = '%s://' % self.__scheme
        auth   = ''

        if self.__username and self.__password:
            auth = '%s:%s@' % (self.__username, urllib.quote(self.__password))
        
        uri = '%s%s%s:%s' % (scheme, auth, self.__hostname, self.__port)
        uri = urllib.basejoin(uri, self.__wservice)

        # Creating the server
        self.__server = xmlrpclib.Server(uri, verbose=self.__verbose)

        del(scheme)
        del(auth)
        del(uri)

        return True

    def run(self):
        """
        Make the connection and start the thread
        """

        while self.__do_run:
            if self.__request_fifo:
                # Getting the function name and arguments from the fifo
                function_name, args = self.__request_fifo.pop()

                # Sending the request
                result, data = self.__send_request(function_name, *args)

                # Sending the result to the main process
                self.emit('xmlrpc-data-received', function_name, result, data)
                
            # Don't eat the CPU
            time.sleep(0.1)

    def __send_request(self, function_name, *args):
        """
        Sends a request to the server
        """
        # Encoding the arguments
        enc_args = []
        for arg in args:

            if type(arg) == dict:
                enc_arg = dump_struct(arg)
            else:
                enc_arg = xmlrpclib.Binary(arg)

            enc_args.append(enc_arg)
        
        # Calling the function
        try:
            function = getattr(self.__server, function_name)
            data     = function(*enc_args)
            result   = True

        except xmlrpclib.Error, error:
            # Xmlrpclib error
            result   = False
            data     = error

            # Getattr failed:
            if hasattr(data, 'faultCode'):
                data.errcode = data.faultCode
                data.errmsg  = data.faultString
                del(data.faultCode)
                del(data.faultString)

        except:
            # Unknown error
            result   = False
            data     = EemyleConnectorError()

        else:
            if self.__verbose:
                sys.stdout.write('\n%s\n' % str(data))

        return (result, data)

    def send_request(self, function_name, *args):
        """
        Stack a request to the internal fifo
        """

        self.__request_fifo.insert(0, [function_name, args])

        return True

    def send_direct_request(self, function_name, *args):
        """
        Sends a request without stacking it
        Directly return the result
        """

        # Sending the request
        return self.__send_request(function_name, *args)

    def stop(self):

        self.__do_run = False


gobject.type_register(EemyleConnector)
gobject.signal_new('xmlrpc-data-received', EemyleConnector, gobject.SIGNAL_RUN_LAST,
                   gobject.TYPE_NONE, [gobject.TYPE_STRING, gobject.TYPE_BOOLEAN, gobject.TYPE_PYOBJECT])

