#!/usr/bin/env python
#
# GEemyle: Terminal module
# Copyright (C) 2003-2006 Easter-eggs <info@easter-eggs.com>
# Authors:
#      Cyril Lacoux (Yack) <clacoux@easter-eggs.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

__author__  = "Cyril Lacoux [Yack] <clacoux@easter-eggs.com>"
__cvsid__   = "$Id$"
__version__ = "$Revision$"[11:-2]
__all__     = [ "EemyleTerminalBox", "ENCODINGS" ]

import pygtk
pygtk.require('2.0')
import gtk
import gobject
import weakref

import vte, pango
import re
import os, signal
import user

import geemyle.config  as config
import geemyle.gettext as gettext

_ = gettext._

ENCODINGS = [
 ('unicode (utf8)', 'UTF-8'),
 ('iso 8859-1',     'ISO-8859-1'),
 ('iso 8859-15',    'ISO-8859-15')
]

WIDTHS = [
 ('automatic size',    0),
 ('1000 columns',   1000),
 ('2000 columns',   2000),
 ('3000 columns',   3000),
 ('5000 columns',   5000),
 ('10000 columns', 10000)
]

terminal_actions = [
  ('TerminalDetach',    gtk.STOCK_ZOOM_FIT,     'D_etach the terminal',    '<control><shift>E',     'Detach the terminal',    'on_terminal_detach_activate'),
  ('TerminalRename',    gtk.STOCK_EDIT,         '_Rename the terminal',    '<control><shift>S',     'Rename the terminal',    'on_terminal_rename_activate'),
  ('TerminalDuplicate', gtk.STOCK_COPY,         '_Duplicate the terminal', '<control><shift>D',     'Duplicate the terminal', 'on_terminal_duplicate_activate'),
]
    
ui_string = """<ui>
  <popup name='TerminalLabelPopup'>
    <menuitem action='TerminalDetach'/>
    <menuitem action='TerminalDuplicate'/>
    <menuitem action='TerminalRename'/>
  </popup>
</ui>"""

# Terminals settings
_all_terminals = weakref.WeakValueDictionary({}) # Class attibute to share it over all terminals
_global_config = {
    'broadcast': False
    }

def fix_actions(actions, instance):
    """
    Helper function to map methods to an instance
    """
    retval = []
    
    for i in range(len(actions)):
        curr = actions[i]
        if len(curr) > 5:
            curr = list(curr)
            curr[5] = getattr(instance, curr[5])
            curr = tuple(curr)
            
        retval.append(curr)
    return retval

def get_gdkcolor_from_tuple(color):
    try:
        return gtk.gdk.Color(*color)

    except:
        return None

def get_tuplecolor_from_gdk(color):
    try:
        red   = color.red
        green = color.green
        blue  = color.blue
        return (red, green, blue)
    except:
        return False

def get_palette(scheme):
    if not config.PALETTE_SCHEMES.has_key(scheme):
        return None

    palette = []
    for color in config.PALETTE_SCHEMES[scheme]:
        color = get_gdkcolor_from_tuple(color)
        if not color:
            return []
        palette.append(color)

    return list(palette)

def load_config():
    """
    Load config for all terminals
    """
    global _all_terminals
    for terminal in _all_terminals.values():
        terminal.load_config()

def get_terminal_broadcast():
    """
    Return the status of terminal broadcast
    """
    global _global_config
    return _global_config['broadcast']

def set_terminal_broadcast(broadcast):
    """
    Send input to all terminals
    """
    global _global_config
    assert broadcast in (False, True)
    _global_config['broadcast'] = broadcast

def is_terminal_opened():
    """
    Returns True if some terminals are opened, False if not
    """
    global _all_terminals
    return len(_all_terminals) != 0
    

class EemyleTerminal(vte.Terminal):
    def __init__(self, session_infos=[], session_commands=[], session_datas={}):

        # Call the parent class
        super(EemyleTerminal, self).__init__()

        # Session infos and commands
        self.session_infos    = session_infos
        self.session_commands = session_commands
        self.session_datas    = session_datas
        self.current_hop      = 0
        self.buffer           = ''
        self.pid              = None

        # Number of columns
        self.col_count = 0

        # Connection
        self.connect_stage  = 0
        self.give_root_pass = False

        # ZModem transfer
        self.zmodem_stage = 0
        self.zmodem_files = []

        # Configuration
        self.__scroll_step = 10
        self.load_config()

        # To feed the child
        self.contentsChangedHandlerId = None

        # To feed orther terminal(s) when this terminal is master
        _all_terminals[id(self)] = self

        # To start the connection after signals connections
        gobject.timeout_add(100, self.__start_connection)

        # XXX ugly hack, but if we set encoding in __start_connection, it just does not work
        gobject.timeout_add(1000, lambda: self.set_encoding("UTF-8"))

    @property
    def all_terminals(self):
        global _global_terminals
        return _all_terminals.values()

    def broadcast_get(self):
        global _global_config
        return _global_config['broadcast']

    def broadcast_set(self, value):
        global _global_config
        assert value in (False, True)
        _global_config['broadcast'] = value
    
    broadcast = property(broadcast_get, broadcast_set)

    def set_font_name(self, fontname):
        self.set_font_from_string(fontname)

    def set_scroll_step(self, scroll_step):
        self.__scroll_step = scroll_step

    def load_config(self):
        # General
        if not config.get('terminal', 'use_system_font'):
            self.set_font_name(config.get('terminal', 'font_name'))
        self.set_audible_bell(config.get('terminal', 'audible_bell'))
        self.set_visible_bell(config.get('terminal', 'visible_bell'))
        self.set_allow_bold(config.get('terminal', 'allow_bold'))
        self.set_cursor_blinks(config.get('terminal', 'cursor_blinks'))
        self.set_mouse_autohide(config.get('terminal', 'mouse_autohide'))
        self.set_word_chars(config.get('terminal', 'word_chars'))

        # Colors
        palette = get_palette(config.get('terminal', 'palette_scheme'))

        if config.COLOR_SCHEMES.has_key(config.get('terminal', 'color_scheme')):
            foreground_color = get_gdkcolor_from_tuple(config.COLOR_SCHEMES[config.get('terminal', 'color_scheme')][0])
            background_color = get_gdkcolor_from_tuple(config.COLOR_SCHEMES[config.get('terminal', 'color_scheme')][1])

        else:
            foreground_color = None
            background_color = None

        if palette:
            if not foreground_color:
                foreground_color = palette[7]
            if not background_color:
                background_color = palette[0]

        if foreground_color and background_color and palette:
            self.set_colors(foreground_color, background_color, palette)
        else:
            if foreground_color:
                self.set_color_foreground(foreground_color)
            if background_color:
                self.set_color_foreground(background_color)
            
        # Effects
        if config.get('terminal', 'background_image_file'):
            self.set_background_image_file(config.get('terminal', 'background_image_file'))
        self.set_background_saturation(config.get('terminal', 'background_saturation'))
        self.set_background_transparent(config.get('terminal', 'background_transparent'))

        # Scrolling
        self.set_scrollback_lines(config.get('terminal', 'scrollback_lines'))
        self.set_scroll_on_output(config.get('terminal', 'scroll_on_output'))
        self.set_scroll_on_keystroke(config.get('terminal', 'scroll_on_keystroke'))
        self.set_scroll_step(config.get('terminal', 'scroll_step'))

        # Compatibility
        self.set_backspace_binding(config.get('terminal', 'backspace_binding'))
        self.set_delete_binding(config.get('terminal', 'delete_binding'))

    def change_font_size(self, direction):
        """
        This method handles increase or decrease font size event
        CONTROL key + mouse scroll
        """

        font_desc = self.get_font()
        font_size = font_desc.get_size()
        w, h      = self.window.get_size()
        
        # Get the new size
        if direction == gtk.gdk.SCROLL_DOWN:
            font_size += pango.SCALE
        else:
            font_size -= pango.SCALE

        # Set the new font
        font_desc.set_size(font_size)
        self.set_font(font_desc)
        
        # Keep the size of the window
        self.set_size_request(w, h)

    def scroll(self, direction, scroll_page=False):
        """
        This method handles mouse scroll events
        """

        adj    = self.get_adjustment()
        value  = adj.get_value()
        upper  = adj.upper
        height = self.get_row_count()

        if not scroll_page:
            if direction == gtk.gdk.SCROLL_DOWN:
                value += self.__scroll_step
            else:
                value -= self.__scroll_step
        else:
            if direction == gtk.gdk.SCROLL_DOWN:
                value += height - 1
            else:
                value -= height - 1

        # Do not scroll outside the terminal
        if value + height > upper:
            value = upper - height
            if value < 0:
                value = 0
        
        adj.set_value(value)

    def __is_selected(self, font, col, row, data):
        return True

    def get_buffer(self):
        return self.get_text(self.__is_selected)

    def on_contents_changed(self, arg):
        buffer = self.get_text(self.__is_selected)

        # Checks if buffer changed
        if buffer == self.buffer:
            return False

        # Don't broadcast any inputs made here
        self.handler_block(self.commitHandlerId)

        self.buffer = buffer

        # Remove all empty lines
        lines = []
        for line in buffer.split('\n'):
            line = line.strip()
            if line:
                lines.append(line)

        if not lines:
            # Reactivating broadcast
            self.handler_unblock(self.commitHandlerId)
            return False

        # Transfer mode
        if self.zmodem_stage:
            m1 = '^zssh >$'
            r1 = re.compile('Sending: (.*)')
            r2 = re.compile('Receiving: (.*)')
            r3 = re.compile('skipped: (.*)')
            m2 = '^Transfer complete$'
            m3 = '^Transfer incomplete$'

            for line in lines:
                # Send the files
                if self.zmodem_stage == 1:
                    if re.match(m1, line):
                        self.zmodem_stage = 2

                        # Reseting the terminal
                        self.reset(True, True)

                        cmd = 'sz -e'
                        for file in self.zmodem_files:
                            cmd += ' "%s"' % file

                        cmd += '\n'
                        self.feed_child(cmd, len(cmd))

                # Read tranfer state
                elif self.zmodem_stage == 2:
                    f1 = r1.findall(line)
                    f2 = r2.findall(line)
                    f3 = r3.findall(line)
                    if f1:
                        # Tranfering files
                        self.emit('new-message', _('Sending file %s...') % f1[0])

                    elif f2:
                        self.emit('new-message', _('Receiving file %s...') % f2[0])

                    elif f3:
                        self.emit('new-message', _('Skipped file %s...') % f3[0])
                        
                    elif re.match(m2, line):
                        # Transfer is finished
                        self.feed_child('\000\000\000', 3)

                        # Tranfering files
                        self.emit('new-message', _('Transfer finished'))

                        # We don't want to go back here
                        self.zmodem_stage = 0
                        #self.disconnect(self.contentsChangedHandlerId)
                        #self.contentsChangedHandlerId = None
                        break

                    elif re.match(m3, line):
                        # Transfer failed
                        self.feed_child('\000\000\000', 3)

                        self.emit('new-message', _('Transfer failed'))

                        # We don't want to go back here
                        self.zmodem_stage = 0
                        #self.disconnect(self.contentsChangedHandlerId)
                        #self.contentsChangedHandlerId = None
                        break


        # SSH internal stuff
        if self.current_hop < len(self.session_infos):

            hostname = self.session_infos[self.current_hop]['server_name']
            ip_addr  = self.session_infos[self.current_hop]['ip']

            # Pubkey
            m1 = '^The authenticity of host \'.* \(.*\)\' can\'t be established.$'
            m2 = '^[RD]SA key fingerprint is [\d\w:]+.$'
            m3 = '^Are you sure you want to continue connecting \(yes/no\)\?$'

            # Connection
            m4 = '^debug1: Connecting to .* \[.*\]'
            m5 = '^debug1: Connection established.$'

            # Authentication
            m6 = '^debug1: Authentication succeeded \(.*\).$'
            m7 = '^debug1: ssh-userauth2 successful: method .*$'            

            pubkey_stage  = 0
            for line in lines:
                # Pubkey
                if pubkey_stage == 0:
                    if re.match(m1, line):
                        pubkey_stage = 1
    
                elif pubkey_stage == 1:
                    if re.match(m2, line):
                        pubkey_stage = 2
    
                elif pubkey_stage == 2:
                    if re.match(m3, line):
                        pubkey_stage = 0
                        self.emit('new-message', _('Accepting public key for %s') % hostname)
                        self.feed_child('yes\n', 4)

                # Connection
                if self.connect_stage == 0:
                    if re.match(m4, line):
                        self.connect_stage = 1
                        self.emit('new-message', _('Connecting to %s...') % hostname)

                elif self.connect_stage == 1:
                    if re.match(m5, line):
                        self.connect_stage = 2
                        self.emit('new-message', _('Connected to %s') % hostname)
                        # Set TERM cols and lines after successful connection
                        self.set_size(self.col_count)

                # Authentication:
                elif self.connect_stage == 2 or self.connect_stage == 3:
                    if self.connect_stage == 2:
                        auth_method = 'pubkey'
                    else:
                        auth_method = 'password'

                    if re.match(m6, line) or re.match(m7, line):
                        self.connect_stage = 0
                        self.current_hop  += 1
                        self.reset(True, True)

                        percent = 100 * self.current_hop / len(self.session_infos)
                        self.emit('set-percent', percent,  _('Connecting...'))
                        self.emit('new-message', _('Authentication succeeded on %s (%s).') % (hostname, auth_method))

                        # We don't want to go back here
                        if self.current_hop == len(self.session_infos):
                            self.connect_stage = 4
                        #    self.disconnect(self.contentsChangedHandlerId)
                        #    self.contentsChangedHandlerId = None

        # Get the last line:
        lines.reverse()
        for line in lines:
            if line:
                break
    
        # Authentication
        if self.connect_stage == 2:
            username = self.session_infos[self.current_hop]['admin_name']
            password = self.session_infos[self.current_hop]['admin_pass']

            m1 = '^%s@%s\'s password:$' % (username, ip_addr)
            m2 = '^Password:$'
                
            if re.match(m1, line) or re.match(m2, line):
                self.connect_stage = 3
                self.emit('new-message', _('Sending %s\'s password to %s...') % (username, hostname))
                self.feed_child('%s\n' % password, len(password) + 1)


        # Root password
        if self.give_root_pass and self.session_infos:
            root_name = self.session_infos[-1]['root_name']
            root_pass = self.session_infos[-1]['root_pass']
            sudo_root = self.session_infos[-1]['sudo_root']
            if re.search("sudo -.", root_pass) or sudo_root == 't':
                root_pass = self.session_infos[-1]['admin_pass']
                m = '^SUDO PASSWORD:$'
            else:
                m = '^(Password|Mot de passe)\s*:\s*$'

            if re.match(m, line):
                # We don't want to go back here
                self.give_root_pass = False
                #self.disconnect(self.contentsChangedHandlerId)
                #self.contentsChangedHandlerId = None
                self.feed_child('%s\n' % root_pass, len(root_pass) + 1)
            del(root_pass)

        # Zmodem detect
        if len(line) == 19:
            l = ''
            for char in line:
                l += '%X' % ord(char)

            if l == 'C28A1118423030303030303030303030303030':
                self.zmodem_stage = 2
                # Entering ZModem tranfer mode
                self.feed_child('\000', 1)

        # Reactivating broadcast
        self.handler_unblock(self.commitHandlerId)

        self.emit('activity')

    def __start_connection(self):
        # To have a look to the child
        self.commitHandlerId = self.connect('commit', self.on_commit)
        self.contentsChangedHandlerId = self.connect('contents-changed', self.on_contents_changed)

        if self.session_infos:
            # Remote command
            if self.session_datas.has_key('remote_command'):
                remote_command = self.session_datas['remote_command']
            else:
                remote_command = False

            # Tunnel
            if self.session_datas.has_key('tunnel') and self.session_datas['tunnel'] == True:
                local_port   = self.session_datas['local_port']
                remote_host  = self.session_datas['remote_host']
                remote_port  = self.session_datas['remote_port']
                allow_remote = self.session_datas['allow_remote']
            else:
                local_port   = None
                remote_port  = None
                remote_port  = -1
                allow_remote = False

            if self.session_datas.has_key('compression') and self.session_datas['compression'] == True:
                compression = True
            else:
                compression = False

            if self.session_datas.has_key('x11forwarding') and self.session_datas['x11forwarding'] == True:
                x11forwarding = True
            else:
                x11forwarding = False

            # Constructing the command line
            n_hops = len(self.session_infos)
            argv   = []
            for session_info in self.session_infos:
                index = self.session_infos.index(session_info)

                # Command
                # The first command need to be zssh to be able to send and receive files
                if not argv:
                    argv.append('zssh')
                else:
                    argv.append('ssh')

                # Port on which ssh is listenning on server
                if not int(session_info['ssh_port']) == 22:
                    argv.append('-p')
                    argv.append(session_info['ssh_port'])

                # Verbose mode
                argv.append('-v')

                # Compression
                if compression or config.get('eemyle', 'use_ssh_compression'):
                    argv.append('-C')

                # X11Forwarding
                if x11forwarding:
                    argv.append('-X')

                # Tunnel
                if local_port and remote_port:
                    # Allowing remote hosts to connect to forwarded port only on the first connection
                    if index == 0 and allow_remote:
                        argv.append('-g')

                    argv.append('-L')
                    if index < n_hops - 1:
                        # Using high port to transport the tunnel
                        argv.append('%d:localhost:%d' % (local_port, local_port))
                    else:
                        # Then finally forward the desired port
                        argv.append('%d:%s:%d' % (local_port, remote_host, remote_port))

                # User and IP address
                argv.append('%s@%s' % (session_info['admin_name'], session_info['ip']))

                # Prepare the comand line for the next hop
                argv.append('-t')
                # Use Agent forwarding
                argv.append('-A')

            # Remove the last (unused) argument
            argv.pop()

            # Remote command
            if remote_command:
                argv.append(remote_command)
        else:
            # No session launch a shell
            self.emit('new-message', '')

            # Remote command
            if self.session_datas.has_key('remote_command'):
                argv = self.session_datas['remote_command'].split()
            else:
                argv = config.get('terminal', 'prefered_shell').split()
            
        # Forking the command
        self.pid = self.fork_command(argv[0], argv, directory=user.home)

    def send_files(self, files):
        """
        Start a ZModem process using zssh to send files
        """

        # Are we aldready sending some files on this terminal?
        if not self.zmodem_stage == 0:
            return False

        # Are we in broadcast mode?
        if self.broadcast:
            return False

        # The connection is not ready
        if not self.current_hop == len(self.session_infos):
            return False

        # Are we waiting to send the root password?
        if self.give_root_pass:
            return False

        # Preparing transfer
        self.zmodem_files = files
        self.zmodem_stage = 1
        #self.contentsChangedHandlerId = self.connect('contents-changed', self.on_contents_changed)

        # Entering ZModem tranfer mode
        self.feed_child('\000', 1)

        return True

    def on_commit(self, terminal, text, text_len):
        """
        If the terminal is master, send all inserted text to it's children
        Locking the child 'commit' signal handler prevents looping
        """

        if not self.broadcast:
            return True

        for terminal in self.all_terminals:
            if terminal is self:
                continue
            terminal.handler_block(terminal.commitHandlerId)
            terminal.feed_child(text, text_len)
            terminal.handler_unblock(terminal.commitHandlerId)

        return True
            
    def on_send_command_activate(self, command):
        
        # Keep the focus!
        self.set_property('has-focus', True)
        self.feed_child('%s\n' % command, len(command) + 1)

    def on_stop_feed_child_activate(self):

        #if self.contentsChangedHandlerId:
            #self.disconnect(self.contentsChangedHandlerId)
            #self.contentsChangedHandlerId = None
            
        self.current_hop == len(self.session_infos)
        self.emit('new-message', '')
        self.emit('set-percent', 100, '')

    def on_become_root_activate(self):
        # If broadcast mode, we need to send the root password for each terminals
        if self.broadcast:
            for terminal in self.all_terminals:
                # Don't broadcast any inputs made here
                terminal.handler_block(terminal.commitHandlerId)

                terminal.give_root_pass = True
                root_name = terminal.session_infos[-1]['root_name']
                root_pass = terminal.session_infos[-1]['root_pass']
                sudo_root = terminal.session_infos[-1]['sudo_root']
                if re.search("sudo -.", root_pass) or sudo_root == 't':
                    terminal.on_send_command_activate('sudo -p "SUDO PASSWORD:" -i')
                elif root_name == 'enable':
                    terminal.on_send_command_activate('enable')
                else:
                    terminal.on_send_command_activate('su -')
                # Reactivating broadcast
                terminal.handler_unblock(terminal.commitHandlerId)
        else:
            self.give_root_pass = True
            root_name = self.session_infos[-1]['root_name']
            root_pass = self.session_infos[-1]['root_pass']
            sudo_root = self.session_infos[-1]['sudo_root']
            if re.search("sudo -.", root_pass) or sudo_root == 't':
                self.on_send_command_activate('sudo -p "SUDO PASSWORD:" -i')
            elif root_name == 'enable':
                self.on_send_command_activate('enable')
            else:
                self.on_send_command_activate('su -')

    def on_close_connection_activate(self):
        if not self.pid:
            return False

        try:
            os.kill(self.pid, signal.SIGHUP)
        except:
            return False

        return True

    def set_size(self, col_count):
        # Storing the number of columns
        self.col_count = col_count

        # Get the initial size
        w, h = self.window.get_size()

        # Change the number of columns
        vte.Terminal.set_size(self, col_count, self.get_row_count())

        # Calculate the new width
        w = col_count * self.get_char_width()

        self.set_size_request(w, h)

        return True

    def on_set_size_request(self, view_port, rectangle):
        
        if not self.col_count:
            # Adjust the size
            w    = rectangle.width - 10
        else:
            # Keep the actual size
            w, h = self.window.get_size()
           
        h = rectangle.height

        # Ajust the height depending the heigth of the container
        self.set_size_request(w, h)

        return True


class EemyleIPMITerminal(EemyleTerminal):

    def on_close_connection_activate(self):

        try:
            self.on_send_command_activate("~.")
	except Exception:
            EemyleTerminal.on_close_connection_activate(self)

class EemyleTerminalLabel(gtk.EventBox):

    def __init__(self, name):

        gtk.EventBox.__init__(self)

        self.icon       = gtk.Image()
        self.label      = gtk.Label(name)
        self.bt_close   = gtk.Button()

        self.construct_frame()
        self.show_all()

    def construct_frame(self):
        """
        Create the widget
        """

        hbox = gtk.HBox()
        hbox.set_spacing(5)

        self.set_icon_from_file(config.APP_ICON_FILEPATH)

        hbox.pack_start(self.icon)
        hbox.pack_start(self.label)

        # The button thanks to http://kazehakase.sourceforge.jp/ ;-)
        self.bt_close.set_relief(gtk.RELIEF_NONE)

        image = gtk.image_new_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
        size  = image.size_request()
        self.bt_close.set_size_request(*size)
        self.bt_close.add(image)

        style = gtk.RcStyle()
        style.xthickness = 0
        style.ythickness = 0
        
        self.bt_close.modify_style(style)

        hbox.pack_start(self.bt_close)

        self.add(hbox)

    def highlight(self):
        """
        Highlight the label to track terminal activity
        """

        text = self.label.get_text()
        self.label.set_markup('<span foreground="#EE0000">{}</span>'.format(text))

    def set_icon_from_file(self, filepath):
        """
        Set the icon
        """

        pixbuf = gtk.gdk.pixbuf_new_from_file(filepath)
        icon = gtk.IconSet(pixbuf)
        self.icon.set_from_icon_set(icon, gtk.ICON_SIZE_MENU)

    def set_text(self, text):
        """
        Change the label text
        """

        self.label.set_text(text)

    def get_text(self):
        """
        Get the label's text
        """

        return self.label.get_text()


class EemyleTerminalBox(gtk.ScrolledWindow):
    def __init__(self, geemyle_ui, name=_('Terminal'), session_infos=[], session_commands=[], session_datas={}):

        gtk.ScrolledWindow.__init__(self)

        self.use_prompt_for_window_title = False
        self.use_prompt_for_tab_label    = False

        self.geemyle_ui = geemyle_ui
        self.orig_name  = name
        if 'IPMI' in session_datas and session_datas['IPMI']:
            self.terminal   = EemyleIPMITerminal(session_infos=session_infos, session_commands=session_commands, session_datas=session_datas)
        else:
            self.terminal   = EemyleTerminal(session_infos=session_infos, session_commands=session_commands, session_datas=session_datas)
        self.label      = EemyleTerminalLabel(name)

        self.ChildExitedHandlerId = None
        self.sizeAllocateHandlerId = None

        # Configuration
        self.load_config()

        self.construct_frame()

    def construct_frame(self):
        # Signals connection
        self.terminal.connect('window-title-changed', self.on_terminal_window_title_changed)
        self.terminal.connect('activity',             self.on_terminal_activity)
        self.terminal.connect('event',                self.on_terminal_event)
        self.terminal.connect('new-message',          self.on_terminal_new_message)
        self.terminal.connect('set-percent',          self.on_terminal_set_percent)

        self.label.connect('event',                   self.on_label_event)
        self.label.bt_close.connect('clicked',        self.on_label_bt_close_clicked)

        self.ChildExitedHandlerId = self.terminal.connect('child-exited', self.on_terminal_child_exited)

        self.set_shadow_type(gtk.SHADOW_NONE)

        # Embedd the terminal into a viewport
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        self.add_with_viewport(self.terminal)

        # The vertical scrollbar must control the terminal adjustment
        vp = self.get_child()
        vp.set_shadow_type(gtk.SHADOW_NONE)
        vp.set_vadjustment(None)

        sb = self.get_vscrollbar()
        sb.set_adjustment(self.terminal.get_adjustment())

        # The height of the terminal must follow the height of the viewport
        self.sizeAllocateHandlerId = vp.connect('size-allocate', self.terminal.on_set_size_request)

        self.show_all()
        
    def load_config(self):

        if config.get('terminal', 'scrollbar_position') == 'left':
            if config.get('terminal', 'show_scrollbar'):
                self.set_placement(gtk.CORNER_TOP_RIGHT)

        elif config.get('terminal', 'scrollbar_position') == 'right':
            if config.get('terminal', 'show_scrollbar'):
                self.set_placement(gtk.CORNER_TOP_LEFT)

        else:
            self.get_vscrollbar().hide_all()

        self.terminal.load_config()

        self.use_prompt_for_window_title = config.get('terminal', 'use_prompt_for_window_title')
        self.use_prompt_for_tab_label    = config.get('terminal', 'use_prompt_for_tab_label')
        self.on_terminal_window_title_changed()

        self.show_all()

    def on_terminal_window_title_changed(self, terminal=None):
        """
        Change the tab label and window title depending of the terminal's window title
        """

        if not terminal:
            terminal = self.terminal

        title = terminal.get_window_title()

        if self.use_prompt_for_window_title and title:
            window_title = '%s - %s' % (title, config.PACKAGE)
        else:
            window_title = '%s - %s' % (self.orig_name, config.PACKAGE)

        if self.use_prompt_for_tab_label and title:
            label_title  = title
        else:
            label_title  = self.orig_name

        self.geemyle_ui.set_title(window_title)
        self.label.set_text(label_title)

    def on_terminal_activity(self, terminal):

        current_page, label, child = self.geemyle_ui.get_current_child()
        if child != self:
            self.label.highlight()

    def on_terminal_event(self, terminal, event):
        """
        Display the popup menu
        """
        if event.type == gtk.gdk.BUTTON_PRESS:
            if event.button == 3:
                menu = self.geemyle_ui.ui.get_widget('/TerminalPopup')
                menu.popup(None, None, None, event.button, event.time)
                return True
            
        # Mouse Scroll
        elif event.type == gtk.gdk.SCROLL:
            state = event.get_state()

            if state & gtk.gdk.CONTROL_MASK:
                terminal.change_font_size(event.direction)

            else:
                if state & gtk.gdk.SHIFT_MASK:
                    scroll_page = True
                else:
                    scroll_page = False

                terminal.scroll(event.direction, scroll_page)

            # Returning True to disable internal VTE_TERMINAL scrolling action
            return True

        return False

    def on_terminal_new_message(self, terminal, message):
        """
        Display a message in status bar
        """

        self.geemyle_ui.on_set_message(message)

    def on_terminal_set_percent(self, terminal, percent, message):
        """
        Display a percentage in status bar
        """

        self.geemyle_ui.on_set_percent(percent, message)

    def on_terminal_child_exited(self, terminal):
        """
        Close the corresponding tab
        """
        global _all_terminals
        self.geemyle_ui.on_terminal_close_activate(child=self)
        # Disabling broadcast if there is no more opened terminals
        if get_terminal_broadcast() and len(_all_terminals) == 1:
            self.geemyle_ui.ui.get_widget('/TerminalToolBar/TerminalBroadcast').set_active(False)
        self.geemyle_ui = None
        self.label.destroy()
        self.label = None

        self.terminal.destroy()
        self.terminal = None

        self.destroy()
        return True

    def create_popup_menu(self):
        
        ag = gtk.ActionGroup('TerminalActions')
        actions = fix_actions(terminal_actions, self.geemyle_ui)
        ag.add_actions(actions, self)
        
        ui = gtk.UIManager()
        ui.insert_action_group(ag, 0)
        ui.add_ui_from_string(ui_string)

        # To move the terminal to another window
        i = 0
        sendto_menu = gtk.Menu()
        for geemyle_ui in self.geemyle_ui.get_all_windows():
            if not geemyle_ui == self.geemyle_ui:
                win_title = geemyle_ui.get_title()
                menu_item = gtk.MenuItem(win_title)
                menu_item.connect('activate', self.geemyle_ui.on_terminal_sendto_activate, self, geemyle_ui)
                sendto_menu.append(menu_item)

                i += 1

        menu_item = gtk.MenuItem(_('_Send to...'))
        if i:
            menu_item.set_submenu(sendto_menu)
        else:
            menu_item.set_sensitive(False)

        popup_menu = ui.get_widget('/TerminalLabelPopup')
        popup_menu.append(menu_item)

        return popup_menu

    def on_label_event(self, terminal_label, event):
        """
        Display the popup menu
        """
        if event.type == gtk.gdk.BUTTON_PRESS:
            if event.button == 3:

                popup_menu = self.create_popup_menu()
                popup_menu.popup(None, self, None, event.button, event.time)
                popup_menu.show_all()
                return True

        return False

    def on_label_bt_close_clicked(self, button):
        """
        Close the terminal
        """
        
        self.terminal.on_close_connection_activate()

    def get_label(self):
        
        return self.label
